/* Code based on Marlin Plugin Manager
 * Copyright (C) 2011 Lucas Baudin <xapantu@gmail.com>
 *
 * Author: Zeeshan Ali (Khattak) <zeeshanak@gnome.org> (from Rygel)
 *
 * This file is part of Marlin.
 *
 * Marlin is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Marlin is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

public class Gazette.PluginManager : GLib.Object {
    public signal void plugin_enabled (Plugin.Base plugin);
    delegate Gazette.Plugin.Base ModuleInitFunc ();
    Gee.HashMap<string,Gazette.Plugin.Base> plugin_hash;
    Gee.HashMap<string,Gazette.Plugin.Base> enabled_plugins;
    Settings settings;
    string settings_field;
    string plugin_dir;
    Gee.HashMap <string,string> names_paths;

    public PluginManager (Settings settings, string field, string plugin_dir) {
        settings_field = field;
        this.settings = settings;
        this.plugin_dir = plugin_dir;
        this.settings.changed.connect(update_plugins);
        plugin_hash = new Gee.HashMap<string,Gazette.Plugin.Base>();
        names_paths = new Gee.HashMap<string,string>();
        enabled_plugins = new Gee.HashMap<string, Gazette.Plugin.Base>();
    }

    public Gee.Collection<Gazette.Plugin.Base> get_plugins () {
        return plugin_hash.values;
    }
    public Gee.Collection<Gazette.Plugin.Base> get_enabled_plugins () {
        return enabled_plugins.values;
    }
    public void load_plugins () {
        load_modules_from_dir(plugin_dir);
    }

    private void load_modules_from_dir (string path) {
        File dir = File.new_for_path(path);

        string attributes = FileAttribute.STANDARD_NAME + "," +
                            FileAttribute.STANDARD_TYPE;

        FileInfo info;
        FileEnumerator enumerator;

        try {
            enumerator = dir.enumerate_children
                                        (attributes,
                                         FileQueryInfoFlags.NONE);

            info = enumerator.next_file ();
            while(info != null) {
                string file_name = info.get_name ();
                string file_path = Path.build_filename (dir.get_path (), file_name);

                if(file_name.has_suffix(".plug")) {
                    load_plugin_keyfile(file_path, dir.get_path ());
                }
                info = enumerator.next_file ();
            }
        }
        catch(Error error) {
            critical ("Error listing contents of folder '%s': %s",
                      dir.get_path (),
                      error.message);
        }
    }

    private void load_module (string file_path) {
        Module module = Module.open (file_path, ModuleFlags.BIND_LOCAL);
        if (module == null) {
            warning ("Failed to load module from path '%s': %s",
                     file_path,
                     Module.error ());
            return;
        }

        void* function;

        if (!module.symbol("plugin_init", out function)) {
            warning ("Failed to find entry point function '%s' in '%s': %s",
                     "plugin_init",
                     file_path,
                     Module.error ());
            return;
        }

        unowned ModuleInitFunc plugin_init = (ModuleInitFunc) function;
        assert (plugin_init != null);

        /* We don't want our modules to ever unload */
        module.make_resident ();
        Gazette.Plugin.Base plug = plugin_init ();

        debug ("Loaded module source: '%s', id: '%s'", module.name(), plug.id);

        if(plug != null) {
            plugin_hash[plug.id] = plug;
            if (plug.id in settings.get_strv (settings_field)) {
                enabled_plugins[plug.id] = plug;
            }
        }
        else {
            critical("A plugin failed to load");
        }
    }

    private void load_plugin_keyfile (string path, string parent) {
        var keyfile = new KeyFile();
        try {
            keyfile.load_from_file(path, KeyFileFlags.NONE);
            string name = keyfile.get_string("Plugin", "Name");
            string _path = keyfile.get_string("Plugin", "File");
            names_paths[name] = _path;
            load_module(Path.build_filename(parent, _path));
        }
        catch(Error e) {
            warning("Couldn't open the keyfile: %s, %s", path, e.message);
        }
    }
    public string [] get_available_plugins () {
        return names_paths.keys.to_array ();
    }

    public void load_module_from_id(string id) {
        if (plugin_hash.has_key(id))
            return;
        if (names_paths.has_key(id)) {
            load_module(Path.build_filename(plugin_dir, names_paths[id]));
        }
    }
    void update_plugins (string field) {
        debug(@"Settings field: $field changed");
        if (field == settings_field) {
            var enabled_plugins = settings.get_strv (settings_field);
            var iterator = plugin_hash.map_iterator ();
            while (iterator.has_next ()) {
                iterator.next ();
                var id = iterator.get_key ();
                if (!(id in enabled_plugins)) {
                    disable_plugin(id);
                    iterator.unset ();
                }
            }
            foreach (var id in enabled_plugins) {
                if (!plugin_hash.has_key(id)) {
                    load_module (plugin_dir + "/" + names_paths[id]);
                    if (plugin_hash[id] != null)
                        plugin_enabled (plugin_hash[id]);
                }
            }
        }
    }
    //FIXME: does not delete the plugin from memory.
    void disable_plugin (string id) {
        debug(@"Disabling plugin. ID: $id");
        var plug = plugin_hash[id];
        plug.destroy ();
    }
}
